# -*- coding: utf-8 -*-
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models, connection
from django.dispatch import receiver

from .signals import m2m_history_changed

try:
    from django.db.transaction import atomic
except ImportError:
    from django.db.transaction import commit_on_success as atomic


class ManyToManyHistoryVersion(models.Model):
    class Meta:
        unique_together = ('content_type', 'object_id', 'field_name', 'time')

    content_type = models.ForeignKey(ContentType, related_name='m2m_history_versions', db_index=True)
    object_id = models.BigIntegerField(db_index=True)
    object = generic.GenericForeignKey('content_type', 'object_id')

    field_name = models.CharField(max_length=50, db_index=True)
    time = models.DateTimeField(db_index=True)

    count = models.PositiveIntegerField(default=0)
    added_count = models.PositiveIntegerField(default=0)
    removed_count = models.PositiveIntegerField(default=0)

    @property
    def m2m(self):
        return getattr(self.object, self.field_name)

    @property
    def prev(self):
        try:
            return self.m2m.versions.filter(time__lt=self.time).order_by('-time')[0]
        except IndexError:
            return None

    @property
    def next(self):
        try:
            return self.m2m.versions.filter(time__gt=self.time).order_by('time')[0]
        except IndexError:
            return None

    def items(self, **kwargs):
        return self.m2m.were_at(self.time, **kwargs)

    def added(self, **kwargs):
        return self.m2m.added_at(self.time, **kwargs)

    def removed(self, **kwargs):
        return self.m2m.removed_at(self.time, **kwargs)

    @atomic
    def delete(self, *args, **kwargs):
        self.delete_version_items()
        super(ManyToManyHistoryVersion, self).delete(*args, **kwargs)

    def delete_version_items(self):
        qs = self.m2m.queryset_through

        next = self.next
        prev = self.prev

        if next is None and prev is None:
            # it's only version -> delete all
            qs.all().delete()
        elif next is None and prev:
            # it's last version
            # delete all current entered items
            qs.filter(time_from=self.time).delete()
            # all left items are not left
            qs.filter(time_to=self.time).update(time_to=None)
        elif next and prev:
            # it's version in the middle
            cursor = connection.cursor()

            # all, who left now and exist in next version -> set right time_to value, which get from last part:
            # 1. make temporary table, because of partial indexes cannot let do all operations in current table
            # 2. remove second part of members, who left in first part
            # 3. update time_to of first part of members using temp table
            # 4. drop temp table
            sql = '''
                CREATE TEMP TABLE m2m_history_items_temp AS
                    SELECT second.%(m2m_object_name)s, second.%(m2m_item_name)s, second.time_from, second.time_to
                    FROM %(m2m_table_name)s AS second
                    WHERE (second.time_from = %(next_time_from)s
                        AND second.%(m2m_object_name)s = %(object_id)s
                        AND second.%(m2m_item_name)s IN (
                            SELECT first.%(m2m_item_name)s FROM %(m2m_table_name)s AS first
                                WHERE (first.%(m2m_object_name)s = second.%(m2m_object_name)s
                                    AND first.time_to = %(time_to)s
                                    AND first.%(m2m_item_name)s IN %(items)s)));

                DELETE FROM %(m2m_table_name)s AS second
                    WHERE (second.time_from = %(next_time_from)s
                        AND second.%(m2m_object_name)s = %(object_id)s
                        AND second.%(m2m_item_name)s IN (
                            SELECT first.%(m2m_item_name)s FROM %(m2m_table_name)s AS first
                                WHERE (first.%(m2m_object_name)s = second.%(m2m_object_name)s
                                    AND first.time_to = %(time_to)s
                                    AND first.%(m2m_item_name)s IN %(items)s)));

                UPDATE %(m2m_table_name)s AS first
                    SET time_to = second.time_to
                    FROM "m2m_history_items_temp" AS second
                    WHERE first.%(m2m_object_name)s = %(object_id)s
                      AND first.time_to = %(time_to)s
                      AND first.%(m2m_item_name)s IN %(items)s
                      AND first.%(m2m_item_name)s = second.%(m2m_item_name)s
                      AND second.%(m2m_object_name)s = first.%(m2m_object_name)s
                      AND second.time_from = %(next_time_from)s;

                DROP TABLE m2m_history_items_temp;
                '''.replace('%(m2m_table_name)s', self.m2m.through._meta.db_table) \
                .replace('%(m2m_object_name)s', self.m2m.source_field_name + '_id') \
                .replace('%(m2m_item_name)s', self.m2m.target_field_name + '_id')

            params = {
                'object_id': self.object.pk,
                'time_to': self.time,
                'items': tuple(next.items(only_pk=True)),
                'next_time_from': next.time
            }
            cursor.execute(sql, params)

            users_next = {'%s_id__in' % self.m2m.target_field_name: next.items(only_pk=True)}

            # all, who left here and not exist in the next -> left in the next
            qs.filter(time_to=self.time).exclude(**users_next).update(time_to=next.time)
            # all, who entered here and not exist in the next -> delete
            qs.filter(time_from=self.time).exclude(**users_next).delete()
            # all, who entered here and exist in the next -> entered in the next
            qs.filter(time_from=self.time).filter(**users_next).update(time_from=next.time)


@receiver(m2m_history_changed)
def save_m2m_history_version(sender, action, instance, reverse, pk_set, field_name, time, **kwargs):
    keep_version = not reverse and instance._meta.get_field(field_name).versions
    if keep_version and action in ['post_add', 'post_remove', 'post_clear'] and len(pk_set):
        defaults = {
            'count': getattr(instance, field_name).get_queryset(only_pk=True).count(),
        }
        if action in ['post_add']:
            defaults['added_count'] = len(pk_set)
        elif action in ['post_remove', 'post_clear']:
            defaults['removed_count'] = len(pk_set)
        version, created = ManyToManyHistoryVersion.objects.get_or_create(
            content_type=ContentType.objects.get_for_model(instance), object_id=instance.pk, field_name=field_name,
            time=time, defaults=defaults)
        if not created:
            version.__dict__.update(defaults)
            version.save()



"""
Improved tests based on Django docs: https://docs.djangoproject.com/en/dev/topics/db/examples/many_to_many/
"""

import time

from django.db import models
from django.test import TestCase
from django.utils import timezone

from .fields import ManyToManyHistoryField
from .models import ManyToManyHistoryVersion


class Publication(models.Model):
    title = models.CharField(max_length=30)


class Article(models.Model):
    headline = models.CharField(max_length=100)
    publications = ManyToManyHistoryField(Publication, versions=True)
    publications_no_versions = ManyToManyHistoryField(Publication, related_name='articles_no_versions')


class ManyToManyHistoryTest(TestCase):

    def assertPublicationsEqual(self, a, b):
        return self.assertListEqual(list(a.order_by('id').values_list('id', flat=True)), sorted([p.id for p in b]))

    def test_m2m_fields_and_methods(self):
        self.assertListEqual(sorted([field.name for field in Article._meta.get_field('publications').rel.through._meta.local_fields]),
                             sorted([u'id', 'time_from', 'time_to', 'article', 'publication']))

    def test_m2m_history_features(self):

        p1 = Publication.objects.create(title='Pub1')
        p2 = Publication.objects.create(title='Pub2')
        p3 = Publication.objects.create(title='Pub3')

        article = Article.objects.create(headline='Article1')
        state_time1 = timezone.now()
        # we need to use sleep here to pass travis mysql tests, becouse Django + mysql doesn't support storing microseconds
        # and as result our state_timeX will be equal
        time.sleep(1)

        article.publications = [p1, p2]
        state_time2 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [p1, p2])
        self.assertEqual(article.publications.count(), 2)
        self.assertEqual(article.publications.through.objects.count(), 2)
        time.sleep(1)

        article.publications = [p3]
        state_time3 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [p3])
        self.assertEqual(article.publications.count(), 1)
        self.assertEqual(article.publications.through.objects.count(), 3)
        time.sleep(1)

        article.publications.add(p2, p1)
        state_time4 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [p1, p2, p3])
        self.assertEqual(article.publications.count(), 3)
        self.assertEqual(article.publications.through.objects.count(), 5)
        time.sleep(1)

        article.publications.remove(p2, p1)
        state_time5 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [p3])
        self.assertEqual(article.publications.count(), 1)
        self.assertEqual(article.publications.through.objects.count(), 5)
        time.sleep(1)

        article.publications = [p1, p2]
        state_time6 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [p1, p2])
        self.assertEqual(article.publications.count(), 2)
        self.assertEqual(article.publications.through.objects.count(), 7)
        time.sleep(1)

        article.publications.clear()
        state_time7 = article.publications.last_update_time()
        self.assertPublicationsEqual(article.publications.all(), [])
        self.assertEqual(article.publications.count(), 0)
        self.assertEqual(article.publications.through.objects.count(), 7)
        time.sleep(1)

        # test of history
        self.assertPublicationsEqual(article.publications.were_at(state_time1), [])
        self.assertPublicationsEqual(article.publications.were_at(state_time2), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_at(state_time3), [p3])
        self.assertPublicationsEqual(article.publications.were_at(state_time4), [p1, p2, p3])
        self.assertPublicationsEqual(article.publications.were_at(state_time5), [p3])
        self.assertPublicationsEqual(article.publications.were_at(state_time6), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_at(state_time7), [])

        # test of added_at
        self.assertPublicationsEqual(article.publications.added_at(state_time2), [p1, p2])
        self.assertPublicationsEqual(article.publications.added_at(state_time3), [p3])
        self.assertPublicationsEqual(article.publications.added_at(state_time4), [p1, p2])
        self.assertPublicationsEqual(article.publications.added_at(state_time5), [])
        self.assertPublicationsEqual(article.publications.added_at(state_time6), [p1, p2])
        self.assertPublicationsEqual(article.publications.added_at(state_time7), [])

        # test of removed_at
        self.assertPublicationsEqual(article.publications.removed_at(state_time2), [])
        self.assertPublicationsEqual(article.publications.removed_at(state_time3), [p1, p2])
        self.assertPublicationsEqual(article.publications.removed_at(state_time4), [])
        self.assertPublicationsEqual(article.publications.removed_at(state_time5), [p1, p2])
        self.assertPublicationsEqual(article.publications.removed_at(state_time6), [p3])
        self.assertPublicationsEqual(article.publications.removed_at(state_time7), [p1, p2])

        # test were_between
        self.assertPublicationsEqual(article.publications.were_between(state_time1, state_time2), [])
        self.assertPublicationsEqual(article.publications.were_between(state_time1, state_time3), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_between(state_time2, state_time3), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_between(state_time2, state_time4), [p1, p2, p3])
        self.assertPublicationsEqual(article.publications.were_between(state_time3, state_time4), [p3])
        self.assertPublicationsEqual(article.publications.were_between(state_time5, state_time6), [p3])
        self.assertPublicationsEqual(article.publications.were_between(state_time5, state_time7), [p1, p2, p3])

        # test added_between
        self.assertPublicationsEqual(article.publications.added_between(state_time2, state_time3), [p1, p2, p3])

        # test removed_between
        self.assertPublicationsEqual(article.publications.removed_between(state_time3, state_time5), [p1, p2])
        self.assertPublicationsEqual(article.publications.removed_between(state_time3, state_time6), [p1, p2, p3])

        # test different arguments
        self.assertListEqual(sorted(list(article.publications.were_at(state_time4, only_pk=True))),
                             sorted(list(map(lambda o: o.pk, article.publications.were_at(state_time4)))))
        with self.assertRaises(ValueError):
            article.publications.were_at(state_time5, unique=False)

        # test versions
        self.assertEqual(ManyToManyHistoryVersion.objects.count(), 6)
        self.assertEqual(article.publications.versions.count(), 6)
        for i in range(2, 8):
            state_time = locals()['state_time%d' % i]
            version = article.publications.versions.get(time=state_time)
            self.assertPublicationsEqual(version.items(),   article.publications.were_at(state_time))
            self.assertPublicationsEqual(version.added(),   article.publications.added_at(state_time))
            self.assertPublicationsEqual(version.removed(), article.publications.removed_at(state_time))
            self.assertEqual(version.count,                 article.publications.were_at(state_time).count())
            self.assertEqual(version.added_count,           article.publications.added_at(state_time).count())
            self.assertEqual(version.removed_count,         article.publications.removed_at(state_time).count())

        # test absence of versions
        article.publications_no_versions = [p1, p2]
        self.assertEqual(ManyToManyHistoryVersion.objects.count(), 6)
        self.assertEqual(article.publications_no_versions.versions.count(), 0)

        # test of deleting last version
        version = article.publications.versions.get(time=state_time7)
        version.delete()
        self.assertEqual(state_time6, article.publications.last_update_time())
        self.assertPublicationsEqual(article.publications.all(), [p1, p2])
        self.assertEqual(article.publications.count(), 2)
        self.assertEqual(article.publications.through.objects.count(), 7)

        # test of deleting version in the middle
        self.assertPublicationsEqual(article.publications.were_at(state_time2), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_at(state_time3), [p3])
        self.assertPublicationsEqual(article.publications.were_at(state_time4), [p1, p2, p3])
        self.assertPublicationsEqual(article.publications.added_at(state_time3), [p3])
        self.assertPublicationsEqual(article.publications.added_at(state_time4), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_between(state_time2, state_time4), [p1, p2, p3])

        version = article.publications.versions.get(time=state_time3)
        version.delete()

        self.assertPublicationsEqual(article.publications.were_at(state_time2), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_at(state_time3), [p1, p2])
        self.assertPublicationsEqual(article.publications.were_at(state_time4), [p1, p2, p3])
        self.assertPublicationsEqual(article.publications.added_at(state_time3), [])
        self.assertPublicationsEqual(article.publications.added_at(state_time4), [p3])
        self.assertPublicationsEqual(article.publications.were_between(state_time2, state_time4), [p1, p2])

    def test_m2m_default_features(self):
        """
        Build-in test from https://docs.djangoproject.com/en/dev/topics/db/examples/many_to_many/
        """
        p1 = Publication(title='The Python Journal')
        p1.save()
        p2 = Publication(title='Science News')
        p2.save()
        p3 = Publication(title='Science Weekly')
        p3.save()

        a1 = Article(headline='Django lets you build Web apps easily')
        a1.save()
        a1.publications.add(p1)

        a2 = Article(headline='NASA uses Python')
        a2.save()
        a2.publications.add(p1, p2)
        a2.publications.add(p3)
        a2.publications.add(p3)
        with self.assertRaises(TypeError):
            a2.publications.add(a1)

        p4 = a2.publications.create(title='Highlights for Children')
        self.assertPublicationsEqual(a1.publications.all(), [p1])
        self.assertPublicationsEqual(a2.publications.all(), [p4, p2, p3, p1])

        self.assertPublicationsEqual(p2.article_set.all(), [a2])
        self.assertPublicationsEqual(p1.article_set.all(), [a1, a2])
        self.assertPublicationsEqual(Publication.objects.get(id=4).article_set.all(), [a2])

        self.assertPublicationsEqual(Article.objects.filter(publications__id=1), [a1, a2])
        self.assertPublicationsEqual(Article.objects.filter(publications__pk=1), [a1, a2])
        self.assertPublicationsEqual(Article.objects.filter(publications=1), [a1, a2])
        self.assertPublicationsEqual(Article.objects.filter(publications=p1), [a1, a2])

        self.assertPublicationsEqual(Article.objects.filter(publications__title__startswith="Science"), [a2, a2])
        self.assertPublicationsEqual(
            Article.objects.filter(publications__title__startswith="Science").distinct(), [a2])
        self.assertEqual(Article.objects.filter(publications__title__startswith="Science").count(), 2)
        self.assertEqual(Article.objects.filter(publications__title__startswith="Science").distinct().count(), 1)

        self.assertPublicationsEqual(Article.objects.filter(publications__in=[1, 2]).distinct(), [a1, a2])
        self.assertPublicationsEqual(Article.objects.filter(publications__in=[p1, p2]).distinct(), [a1, a2])

        self.assertPublicationsEqual(Publication.objects.filter(id=1), [p1])
        self.assertPublicationsEqual(Publication.objects.filter(pk=1), [p1])
        self.assertPublicationsEqual(
            Publication.objects.filter(article__headline__startswith="NASA"), [p4, p3, p2, p1])

        self.assertPublicationsEqual(Publication.objects.filter(article__id=1), [p1])
        self.assertPublicationsEqual(Publication.objects.filter(article__pk=1), [p1])
        self.assertPublicationsEqual(Publication.objects.filter(article=1), [p1])
        self.assertPublicationsEqual(Publication.objects.filter(article=a1), [p1])

        self.assertPublicationsEqual(Publication.objects.filter(article__in=[1, 2]).distinct(), [p1, p2, p3, p4])
        self.assertPublicationsEqual(Publication.objects.filter(article__in=[a1, a2]).distinct(), [p1, p2, p3, p4])
        self.assertPublicationsEqual(Article.objects.exclude(publications=p2), [a1])

        p1.delete()
        self.assertPublicationsEqual(Publication.objects.all(), [p2, p3, p4])
        a1 = Article.objects.get(pk=1)
        self.assertPublicationsEqual(a1.publications.all(), [])

        a2.delete()
        self.assertPublicationsEqual(Article.objects.all(), [a1])
        self.assertPublicationsEqual(p2.article_set.all(), [])

        a4 = Article(headline='NASA finds intelligent life on Earth')
        a4.save()
        p2.article_set.add(a4)
        p2.article_set.all()
        self.assertPublicationsEqual(p2.article_set.all(), [a4])
        self.assertPublicationsEqual(a4.publications.all(), [p2])

        new_article = p2.article_set.create(headline='Oxygen-free diet works wonders')
        self.assertPublicationsEqual(p2.article_set.all(), [a4, new_article])
        a5 = p2.article_set.all()[1]
        self.assertPublicationsEqual(a5.publications.all(), [p2])

        # Removing Publication from an Article:
        self.assertPublicationsEqual(p2.article_set.all(), [a4, a5])
        a4.publications.remove(p2)
        self.assertPublicationsEqual(p2.article_set.all(), [a5])
        self.assertPublicationsEqual(a4.publications.all(), [])

        # And from the other end:
        p2.article_set.remove(a5)
        self.assertPublicationsEqual(p2.article_set.all(), [])
        self.assertPublicationsEqual(a5.publications.all(), [])

        # Relation sets can be assigned. Assignment clears any existing set members:
        self.assertPublicationsEqual(a4.publications.all(), [])
        a4.publications = [p3]
        self.assertPublicationsEqual(a4.publications.all(), [p3])

        # Relation sets can be cleared:
        p2.article_set.clear()
        self.assertPublicationsEqual(p2.article_set.all(), [])

        # And you can clear from the other end:
        p2.article_set.add(a4, a5)
        self.assertPublicationsEqual(p2.article_set.all(), [a4, a5])
        self.assertPublicationsEqual(a4.publications.all(), [p2, p3])
        a4.publications.clear()
        self.assertPublicationsEqual(a4.publications.all(), [])
        self.assertPublicationsEqual(p2.article_set.all(), [a5])

        # Recreate the Article and Publication we have deleted:
        p1 = Publication(title='The Python Journal')
        p1.save()
        a2 = Article(headline='NASA uses Python')
        a2.save()
        a2.publications.add(p1, p2, p3)

        # Bulk delete some Publications - references to deleted publications should go:
        Publication.objects.filter(title__startswith='Science').delete()
        self.assertPublicationsEqual(Publication.objects.all(), [p4, p1])
        self.assertPublicationsEqual(Article.objects.all(), [a1, a2, a4, a5])
        self.assertPublicationsEqual(a2.publications.all(), [p1])

        # Bulk delete some articles - references to deleted objects should go:
        q = Article.objects.filter(headline__startswith='Django')
        self.assertPublicationsEqual(q, [a1])
        q.delete()

        # After the delete(), the QuerySet versions needs to be cleared, and the referenced objects should be gone:
        self.assertPublicationsEqual(q, [])
        self.assertPublicationsEqual(p1.article_set.all(), [a2])

        # An alternate to calling clear() is to assign the empty set:
        p1.article_set = []
        self.assertPublicationsEqual(p1.article_set.all(), [])

        a2.publications = [p1, p4]
        self.assertPublicationsEqual(a2.publications.all(), [p1, p4])
        a2.publications = []
        self.assertPublicationsEqual(a2.publications.all(), [])
